
/*
 *  dus is a simple utility designed to display the list of files and
 *  directories with disk occupancy, sorted according to it.
 *
 *  Copyright (c) 2009 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of dus.
 *
 *  dus is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  dus is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dus.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>

#define BUFFER_SIZE 1024

typedef int64_t size_sum_t;

size_sum_t file_or_dir_size(char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat dummy;
  size_sum_t result;
  char subname[BUFFER_SIZE];

  result = 0;

  if(lstat(name, &dummy) != 0) {
    printf("Can not stat %s: %s\n", name, strerror(errno));
    exit (1);
  }

  if(S_ISLNK(dummy.st_mode)) {
    return 0;
  }

  dir = opendir(name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(strcmp(dir_e->d_name, ".") &&
         strcmp(dir_e->d_name, "..")) {
        snprintf(subname, BUFFER_SIZE, "%s/%s", name, dir_e->d_name);
        result += file_or_dir_size(subname);
      }
    }
    closedir(dir);
  } else {
    if(S_ISREG(dummy.st_mode)) {
      /* printf("%d %s\n", dummy.st_size, name); */
      result += dummy.st_size;
    }
  }

  return result;
}

/**********************************************************************/

struct file_with_size {
  char *filename;
  size_sum_t size;
  struct file_with_size *next;
};

struct file_with_size *create(char *name, struct file_with_size *current) {
  struct file_with_size *result;
  result = malloc(sizeof(struct file_with_size));
  result->filename = strdup(name);
  result->size = file_or_dir_size(name);
  result->next = current;
  return result;
}

void destroy(struct file_with_size *node) {
  struct file_with_size *next;
  while(node) {
    next = node->next;
    free(node->filename);
    free(node);
    node = next;
  }
}

/**********************************************************************/

int compare_files(const void *x1, const void *x2) {
  const struct file_with_size **f1, **f2;

  f1 = (const struct file_with_size **) x1;
  f2 = (const struct file_with_size **) x2;

  if((*f1)->size < (*f2)->size) {
    return -1;
  } else if((*f1)->size > (*f2)->size) {
    return 1;
  } else {
    return 0;
  }
}


void print_sorted(struct file_with_size *root, int height) {
  struct file_with_size *node;
  struct file_with_size **nodes;
  int nb, n, first;

  nb = 0;
  for(node = root; node; node = node->next) {
    nb++;
  }

  nodes = malloc(nb * sizeof(struct file_with_size *));

  n = 0;
  for(node = root; node; node = node->next) {
    nodes[n++] = node;
  }

  qsort(nodes, nb, sizeof(struct file_with_size *), compare_files);

  /*
  for(n = 0; n < nb; n++) {
    printf("%lld %s\n",
           nodes[n]->size,
           nodes[n]->filename);
  }
  */

  first = 0;
  if(height > 0 && height < nb) {
    first = nb - height;
  }
  for(n = first; n < nb; n++) {
    if(nodes[n]->size < 1024) {
      printf("% 7d %s\n",
             ((int) nodes[n]->size),
             nodes[n]->filename);
    } else if(nodes[n]->size < 1024 * 1024) {
      printf("% 6.1fK %s\n",
             ((double) (nodes[n]->size))/(1024.0),
             nodes[n]->filename);
    } else if(nodes[n]->size < 1024 * 1024 * 1024) {
      printf("% 6.1fM %s\n",
             ((double) (nodes[n]->size))/(1024.0 * 1024),
             nodes[n]->filename);
    } else {
      printf("% 6.1fG %s\n",
             ((double) (nodes[n]->size))/(1024.0 * 1024.0 * 1024.0),
             nodes[n]->filename);
    }
  }

  free(nodes);
}

/**********************************************************************/

int main(int argc, char **argv) {
  int k;
  struct file_with_size *root;

  root = 0;

  if(argc > 1) {
    for(k = 1; k < argc; k++) {
      root = create(argv[k], root);
    }
  } else {
    DIR *dir;
    struct dirent *dir_e;
    dir = opendir(".");
    if(dir) {
      while((dir_e = readdir(dir))) {
        if(strcmp(dir_e->d_name, ".") &&
           strcmp(dir_e->d_name, "..")) {
          root = create(dir_e->d_name, root);
        }
      }
      closedir(dir);
    }
  }

  if(isatty(STDOUT_FILENO)) {
    struct winsize win;
    if(ioctl (STDOUT_FILENO, TIOCGWINSZ, (char *) &win)) {
      printf("Can not get the tty size: %s\n", strerror(errno));
      exit (1);
    }
    print_sorted(root, win.ws_row - 2);
  } else {
    print_sorted(root, -1);
  }

  destroy(root);

  exit(0);
}
